#! /usr/bin/env python
"""
Author: Ethan Payne (ethan.payne@ligo.org)

Calibration marginalization of any generic set of samples. Note that this needs to be used within reason.
The calibration marginalization generates N_ifo matrices of size N_freq_bins by N_calibration_curves.
This can be very memory intensive.

We assume that the likelihood required is only the GWTransientLikelihood with and without calibration marginalization.
No other marginalizations are included. The calibration marginalized likelihood can work with both distance and phase
marginalization if desired.

Also note that the rejection sampling used can be quite lossy - even though it does not significantly change the
posterior distribution.

Inputs
------
posterior_sample_file: Bilby result .json or .dat file
    Either a Bilby result file or a .dat file with headers. Contains the posterior samples of the distribution to be
    used in the reweighting calculation.
data_dump_file: .pickle file
    File produced from bilby_pipe_generation config.ini. This gives an easy way to import in all the required data, and
    in the case of bilby_pipe runs, ensure the same data is used between the original PE and the reweighting.
number_of_calibration_curves: int
    Number of draws of the calibration model which we marginalize over.
    Default: 100.
reevaluate_likelihood: bool
    Sets whether we want to also reevaluate the likelihood without calibration uncertainty. This can ensure that the
    likelihood definitions are consistent (i.e. log_likelihood or log_likelihood - log_noise_evidence).
    Default: True

Output
------
reweighted_posterior_samples.dat:
    Data file with the rejection sampled posterior samples from the supplied posterior file
weights.dat:
    File with the weight for each sample in the supplied posterior file. This allows the user to plot results using the
    the weights as opposed to the rejection sampling posterior samples.
"""

import os
import sys

import argparse
from copy import deepcopy, copy

import numpy as np
import pandas as pd
import pickle

import bilby
import bilby_pipe

# So I can import the RIFT source while not making this some setup.py-able package
# TODO this should not be a hardcoded path!

import RIFT.calmarg.rift_source as rift_source

from  bilby.core.utils import logger


# Alternate implementation of bilby.core.result.reweight
#   - we are using draws, not the prior!
#   - sometimes we get delta-function defailt cal priors, so we get infinities//nan
def alt_reweight(result, label=None, new_likelihood=None, new_prior=None,
             old_likelihood=None, old_prior=None, conversion_function=None, npool=1,
             verbose_output=False, resume_file=None, n_checkpoint=5000,
             use_nested_samples=False):
    """ Reweight a result to a new likelihood/prior using rejection sampling

    Parameters
    ==========
    label: str, optional
        An updated label to apply to the result object
    new_likelihood: bilby.core.likelood.Likelihood, (optional)
        If given, the new likelihood to reweight too. If not given, likelihood
        reweighting is not applied
    new_prior: bilby.core.prior.PriorDict, (optional)
        If given, the new prior to reweight too. If not given, prior
        reweighting is not applied
    old_likelihood: bilby.core.likelihood.Likelihood, (optional)
        If given, calculate the old likelihoods from this object. If not given,
        the values stored in the posterior are used.
    old_prior: bilby.core.prior.PriorDict, (optional)
        If given, calculate the old prior from this object. If not given,
        the values stored in the posterior are used.
    conversion_function: function, optional
        Function which adds in extra parameters to the data frame,
        should take the data_frame, likelihood and prior as arguments.
    npool: int, optional
        Number of threads with which to execute the conversion function
    verbose_output: bool, optional
        Flag determining whether the weight array and associated prior and
        likelihood evaluations are output as well as the result file
    resume_file: string, optional
        filepath for the resume file which stores the weights
    n_checkpoint: int, optional
        Number of samples to reweight before writing a resume file
    use_nested_samples: bool, optional
        If true reweight the nested samples instead. This can greatly improve reweighting efficiency, especially if the
        target distribution has support beyond the proposal posterior distribution.

    Returns
    =======
    result: bilby.core.result.Result
        A copy of the result object with a reweighted posterior
    new_log_likelihood_array: array, optional (if verbose_output=True)
        An array of the natural-log likelihoods from the new likelihood
    new_log_prior_array: array, optional (if verbose_output=True)
        An array of the natural-log priors from the new likelihood
    old_log_likelihood_array: array, optional (if verbose_output=True)
        An array of the natural-log likelihoods from the old likelihood
    old_log_prior_array: array, optional (if verbose_output=True)
        An array of the natural-log priors from the old likelihood

    """
    from scipy.special import logsumexp

    result = copy(result)

    if use_nested_samples:
        result.posterior = result.nested_samples

    nposterior = len(result.posterior)
    logger.info("Reweighting posterior with {} samples".format(nposterior))

    ln_weights, new_log_likelihood_array, new_log_prior_array, old_log_likelihood_array, old_log_prior_array =\
        bilby.core.result.get_weights_for_reweighting(
            result, new_likelihood=new_likelihood, new_prior=new_prior,
            old_likelihood=old_likelihood, old_prior=old_prior,
            resume_file=resume_file, n_checkpoint=n_checkpoint, npool=npool)

    # ONLY USING LIKEILHOOD: we are NOT changing priors, just using cal draws
    
    ln_weights = new_log_likelihood_array - old_log_likelihood_array

    weights = np.exp(ln_weights)

    # Overwrite the likelihood and prior evaluations
    result.posterior["log_likelihood"] = new_log_likelihood_array
    result.posterior["log_prior"] = new_log_prior_array

    result.posterior = bilby.core.result.rejection_sample(result.posterior, weights=weights)
    result.posterior = result.posterior.reset_index(drop=True)
    logger.info("Rejection sampling resulted in {} samples".format(len(result.posterior)))
    result.meta_data["reweighted_using_rejection_sampling"] = True

    if use_nested_samples:
        result.log_evidence += np.log(np.sum(weights))
    else:
        result.log_evidence += logsumexp(ln_weights) - np.log(nposterior)

    if new_prior is not None:
        for key, prior in new_prior.items():
            result.priors[key] = prior

    if conversion_function is not None:
        data_frame = result.posterior
        if "npool" in inspect.signature(conversion_function).parameters:
            data_frame = conversion_function(data_frame, new_likelihood, new_prior, npool=npool)
        else:
            data_frame = conversion_function(data_frame, new_likelihood, new_prior)
        result.posterior = data_frame

    if label:
        result.label = label
    else:
        result.label += "_reweighted"

    if verbose_output:
        return result, weights, new_log_likelihood_array, \
            new_log_prior_array, old_log_likelihood_array, old_log_prior_array
    else:
        return result



parser = argparse.ArgumentParser(description='calibration marginalization via reweighting of posterior samples')
parser.add(
    "--posterior_sample_file", default=None,
    help="Bilby result file or text file with posterior samples to reweight")
parser.add(
    "--data_dump_file", default=None,
    help="Bilby pipe data_dump file. This stores all the relevant data for PE")
parser.add(
    "--waveform_approximant", default=None,
    help="Approximant. Override what is in the pickle file (as expected if running across settings)")
parser.add(
    "--number_of_calibration_curves", type=int, default=1000,
    help="Number of calibration curve realizations to use to reweight samples.")
parser.add(
    "--time_marginalization", type=bool, default=False,
    help="Uses time maginalization when evaluating the likelihoods. Defaults to False")
parser.add(
    "--phase_marginalization", type=bool, default=False,
    help="Uses phase maginalization when evaluating the likelihoods. Defaults to False")
parser.add(
    "--reevaluate_likelihood", type=bool, default=True,
    help="Reevaluates the likelihood of the given samples using the likelihood from the data_dump.")
parser.add(
    "--use_nested_samples", type=bool, default=False,
    help="Use nested samples with their weights to generate the samples. This significantly improves efficiency.")
parser.add(
    "--data-integration-window-half", default=0.1, type=float,
    help="Half-Interval over which the time marginalization is undertaken")
parser.add(
    "--use_rift_samples", type=bool, default=False,
    help="Uses a different waveform function if using RIFT samples. This matches the phase definitions in RIFT")
parser.add(
    "--h_method", type=str, default='hlmoft',
    help="For RIFT, uses two different options for h(t) reconstruction ")
parser.add("--internal-waveform-fd-L-frame",action='store_true',help='If true, passes extra_waveform_kwargs = {fd_L_frame=True} to lalsimutils hlmoft. Impacts outputs of ChooseFDWaveform calls only.')
parser.add("--internal-waveform-fd-no-condition",action='store_true',help='If true, adds extra_waveform_kwargs = {no_condition=True} to lalsimutils hlmoft. Impacts outputs of ChooseFDWaveform calls only. Provided to enable controlled tests of conditioning impact on PE')
parser.add("--use-gwsignal",default=False,action='store_true',help='Use gwsignal. In this case the approx name is passed as a string to the lalsimulation.gwsignal interface')
parser.add("--use-gwsignal-lmax-nyquist",default=None,type=int,help='Passes lmax_nyquist integer to the gwsignal waveform interface')
parser.add("--fmin", default=None, type=float)
parser.add("--l-max", default=4, type=int)
parser.add("--start_index", default=None, type=int)
parser.add("--end_index", default=None, type=int)
parser.add("--internal-use-normal-reweight", default=None, type=float)
args = parser.parse_args()

with open(args.data_dump_file, "rb") as data_file:
    data = pickle.load(data_file)

start_index = args.start_index
end_index = args.end_index

# read in the posterior samples for reweighting
if args.posterior_sample_file.split(".")[-1] == 'json':
    result = bilby.core.result.read_in_result(args.posterior_sample_file)
    if end_index is not None:
        if end_index > len(result.posterior):
            end_index = len(result.posterior)
    result.posterior = result.posterior.iloc[start_index:end_index].reset_index(drop=True)
elif (args.posterior_sample_file.split(".")[-1] == 'txt') or (args.posterior_sample_file.split(".")[-1] == 'dat'):
    result = bilby.core.result.Result()
    result.posterior = pd.DataFrame(np.genfromtxt(args.posterior_sample_file, names=True))
    if end_index is not None:
        if end_index > len(result.posterior):
            end_index = len(result.posterior)
    result.posterior = result.posterior.iloc[start_index:end_index].reset_index(drop=True)
    result.meta_data = {}

    if args.use_rift_samples:
        result.posterior  = result.posterior.drop(columns=['lnL','ps'])
        result.posterior['p'] = np.log(result.posterior['p'])  # not sure if used, but if so define correctly
        key_swap_dict = {'m1':'mass_1', 'm2':'mass_2', 'a1x':'spin_1x', 'a1y':'spin_1y', 'a1z':'spin_1z',
            'a2x':'spin_2x', 'a2y':'spin_2y', 'a2z':'spin_2z', 'incl':'iota', 'time':'geocent_time',
        'phiorb':'phase', 'p':'log_prior', 'distance':'luminosity_distance', 'lambda1':'lambda_1', 'lambda2':'lambda_2'}

        names = list(result.posterior.keys())  # dangerous to have iterator tied to changing structure
        for old_key in names:
            if old_key in key_swap_dict:
                result.posterior[key_swap_dict[old_key]] = result.posterior[old_key]
                del result.posterior[old_key]
        # add tides if not present
        if not('lambda_1' in result.posterior):
            print(" Populating empty tidal params to avoid hang")
            if end_index:
              result.posterior['lambda_1'] = np.zeros(end_index-start_index)
              result.posterior['lambda_2'] = np.zeros(end_index-start_index)
            else:
              result.posterior['lambda_1'] = np.zeros(len(result.posterior['mass_1']))
              result.posterior['lambda_2'] = np.zeros(len(result.posterior['mass_1']))

outdir = os.path.dirname(os.path.abspath(args.posterior_sample_file))

ifos = data.interferometers
time_marginalization_interval = args.data_integration_window_half #args.time_marginalization_interval

spline_calibration_envelope_dict = bilby_pipe.utils.convert_string_to_dict(
                data.meta_data['command_line_args']['spline_calibration_envelope_dict'])
ifos_for_reweighting = deepcopy(ifos)
for ifo in ifos: # removes any model for the calibration that was set up in the file
    ifo.calibration_model = bilby.gw.calibration.Recalibrate()

if args.data_integration_window_half: #args.time_marginalization:
    result.posterior['geocent_time'] = ifos.start_time * np.ones(len(result.posterior))
    if 'time_jitter' not in result.posterior.keys():
        result.posterior['time_jitter'] = np.zeros(len(result.posterior))

if args.phase_marginalization:
    if not args.use_rift_samples:
        result.posterior['phase'] = np.zeros(len(result.posterior))
if args.use_gwsignal:
    args.h_method = 'gws_hlmoft' 

# Setting up the waveform generator using the data dump features
waveform_arguments = dict(
    reference_frequency=data.meta_data['command_line_args']['reference_frequency'],
    minimum_frequency=args.fmin,
    waveform_approximant=data.meta_data['command_line_args']['waveform_approximant'],
    sampling_frequency=ifos.sampling_frequency,
    h_method=args.h_method)

extra_waveform_kwargs={}
if args.internal_waveform_fd_L_frame:
    extra_waveform_kwargs = {'fd_L_frame':True}
if args.internal_waveform_fd_no_condition:
    extra_waveform_kwargs['no_condition'] = True
if args.use_gwsignal_lmax_nyquist:
    extra_waveform_kwargs['lmax_nyquist'] = int(args.use_gwsignal_lmax_nyquist)
waveform_arguments['extra_waveform_kwargs'] = extra_waveform_kwargs
if args.waveform_approximant:
    waveform_arguments['waveform_approximant'] = args.waveform_approximant

if args.use_rift_samples:
    waveform_arguments['Lmax'] = args.l_max
#    waveform_arguments['waveform_approximant'] = 'SEOBNRv4PHM'
    wf_func = rift_source.RIFT_lal_binary_black_hole
else:
    wf_func = eval('bilby.gw.source.'+data.meta_data['command_line_args']['frequency_domain_source_model'])

waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            frequency_domain_source_model=wf_func,
            sampling_frequency=ifos.sampling_frequency,
            duration=ifos.duration,
            start_time=ifos.start_time,
            waveform_arguments=waveform_arguments)

# Setting up the calibration priors given the spline files for each interferometer
# This is used to generate draws of the calibration uncertainty from their prior
priors = bilby.core.prior.PriorDict()
for ifo in ifos_for_reweighting:
    calibration_file_path = f'{spline_calibration_envelope_dict[ifo.name]}'
    ifo_calibration_priors = bilby.gw.prior.CalibrationPriorDict.from_envelope_file(
        calibration_file_path, ifo.minimum_frequency, ifo.maximum_frequency, 10, ifo.name)

    # TODO FOR DEBUGGING PURPOSES
    # for key in ifo_calibration_priors.keys():
    #     if 'amp' in key:
    #         ifo_calibration_priors[key] = bilby.core.prior.DeltaFunction(peak=0)
    #     elif 'phase' in key:
    #         ifo_calibration_priors[key] = bilby.core.prior.DeltaFunction(peak=0)

    priors.update(ifo_calibration_priors)

# Documentation: priors used for cal marg!  
#   Important sanity check
print(" CALMARG PRIORS")
for name in priors:
    print(name, priors[name])


marg_priors = bilby.core.prior.PriorDict() # priors for the non-calibration likelihood


if args.data_integration_window_half: # args.time_marginalization:
    priors['geocent_time'] = bilby.core.prior.Uniform(
        float(data.meta_data['command_line_args']['trigger_time']) - time_marginalization_interval,
        float(data.meta_data['command_line_args']['trigger_time']) + time_marginalization_interval)

    marg_priors['geocent_time'] = bilby.core.prior.Uniform(
        float(data.meta_data['command_line_args']['trigger_time']) - time_marginalization_interval,
        float(data.meta_data['command_line_args']['trigger_time']) + time_marginalization_interval)

if args.phase_marginalization:
    priors['phase'] = bilby.core.prior.Uniform(0, 2 * np.pi)
    marg_priors['phase'] = bilby.core.prior.Uniform(0, 2 * np.pi)

# Setting up the output for the calibration draw files
calibration_lookup_table = {}
for ifo in ifos:
    calibration_lookup_table[ifo.name] =\
                        f'{outdir}/{ifo.name}_calibration_file.h5'

# WARNING: time marginalization is done with *bool*, so actual value of time window  is NOT USED.
original_likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator,
    time_marginalization=args.data_integration_window_half,
    phase_marginalization=args.phase_marginalization, priors=marg_priors)

calibration_likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos_for_reweighting, waveform_generator=copy(waveform_generator),
    number_of_response_curves=args.number_of_calibration_curves, calibration_marginalization=True,
    priors=priors, time_marginalization=args.data_integration_window_half, phase_marginalization=args.phase_marginalization,
    calibration_lookup_table=calibration_lookup_table)

print(f'Log noise evidence: {original_likelihood.noise_log_likelihood()}')

# TODO have not yet implemented the parameter reconstruction in the case where
# time and calibration marginalization are used
if args.data_integration_window_half: # time_marginalization:
    conversion_function = None
else:
    conversion_function = bilby.gw.conversion.generate_all_bbh_parameters

# Setting up the resume and weight file names
resume_file = None
if (start_index != None) and (start_index != None):
    weights_file = f'{outdir}/weight_files/weights_s{start_index}e{end_index}.dat'
elif start_index != None:
    weights_file = f'{outdir}/weight_files/weights_s{start_index}eNone.dat'
elif end_index != None:
    weights_file = f'{outdir}/weight_files/weights_s0e{end_index}.dat'
else:
    resume_file = f'{outdir}/reweighting_resume.dat'
    weights_file = f'{outdir}/weights.dat'

reweight_func = alt_reweight
if args.internal_use_normal_reweight:
    reweight_func = bilby.core.result.reweight

if args.reevaluate_likelihood:
    result_reweighted, weights, _, _, _, _ = reweight_func(
        result, new_likelihood=calibration_likelihood, old_likelihood=original_likelihood,
        conversion_function=conversion_function, verbose_output=True,
        resume_file=resume_file, n_checkpoint=5000,
        use_nested_samples=args.use_nested_samples)
else:
    result_reweighted, weights, _, _, _, _ = reweight_func(
        result, new_likelihood=conversion_function,
        conversion_function=bilby.gw.conversion.generate_all_bbh_parameters, verbose_output=True,
        resume_file=resume_file, n_checkpoint=5000,
        use_nested_samples=args.use_nested_samples)

# Save the weights to a file, with the sample index in the name
if not os.path.exists(f'{outdir}/weight_files/'):
    os.makedirs(f'{outdir}/weight_files/')

np.savetxt(weights_file, weights)

if (start_index == None) and (end_index == None):
    result_reweighted.save_posterior_samples(filename=outdir+'/reweighted_posterior_samples.dat', outdir=outdir)



#! /usr/bin/env python
#
# GOAL
#   test convert_waveform_coordinates against


import numpy as np
import RIFT.lalsimutils as lalsimutils

import optparse
parser = optparse.OptionParser()
parser.add_option("--npts",type=int,default=3)
parser.add_option("--as-test",action='store_true')
parser.add_option("--verbose",action='store_true')
opts, args = parser.parse_args()


npts=opts.npts

# Generate symthetic injections
P_list =[]
for indx in np.arange(npts):
    P = lalsimutils.ChooseWaveformParams()
    P.randomize()
    P_list.append(P)
npts = len(P_list)


# Cartesian test 0
coord_names=['mc','delta_mc', 'xi', 'chiMinus']
low_level_coord_names=['mc','eta','s1z','s2z'] # assume this is the underlying.  This setup is very common

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

print("Cartesian aligned test 0", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")


# Cartesian test 1
coord_names=['mc','delta_mc','xi','chiMinus']
low_level_coord_names=['mc','delta_mc','s1z','s2z'] # assume this is the underlying.  This setup is very common

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

print("Cartesian aligned test 1", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")

# Cartesian aligned test 2: NOT USED CONVERSION, falls back to default converter
# coord_names=['mc','delta_mc','xi','chiMinus']
# low_level_coord_names=['mc','eta','s1z','s2z'] # assume this is the underlying.  This setup is very common

# x1 = np.zeros((npts,len(coord_names)))
# x2 = np.zeros((npts,len(coord_names)))
# y2 = np.zeros((npts,len(low_level_coord_names)))

# for indx in np.arange(npts):
#     P = P_list[indx]
#     for indx_name  in np.arange(len(coord_names)):
#         x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
#     for indx_name2 in np.arange(len(low_level_coord_names)):
#         y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

# x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
# print(x2)

# print("Cartesian aligned test 2", np.max(np.abs(x1 - x2)))
# err = np.max(np.abs(x1 - x2))
# if opts.as_test and err > 1e-9:
#     raise ValueError(" Large deviation seen ")

# Cartesian precessing test A: chi_p etc
coord_names=['mc','delta_mc','xi','chiMinus','chi_p']
low_level_coord_names=['mc','delta_mc','s1z','s2z','s1x','s1y','s2x', 's2y'] # assume this is the underlying

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

err = np.max(np.abs(x1 - x2))
print("Cartesian precessing test A", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")

# Cartesian test 2b: chi_p
coord_names=['mc','delta_mc','xi','chiMinus','chi_p']
low_level_coord_names=['mc','delta_mc','s1z','s2z','s1x','s1y','s2x', 's2y'] # assume this is the underlying

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

err = np.max(np.abs(x1 - x2))
print("Cartesian precessing test B", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")


# Cartesian test 3
coord_names=['mu1','mu2','delta_mc','chiMinus']
low_level_coord_names=['mc','delta_mc','s1z','s2z'] # assume this is the underlying

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

print("Cartesian aligned test 3", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")


# Polar test
coord_names=['mc','delta_mc','xi','chiMinus','s1x','s1y', 's2x', 's2y']
low_level_coord_names=['mc','delta_mc','chi1','cos_theta1', 'phi1','chi2', 'cos_theta2', 'phi2'] # assume this is the underlying


x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

print("Polar test 1", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")


# Cylinder pseudo-polar
coord_names=['mc','delta_mc','xi','chiMinus','s1x','s1y', 's2x', 's2y', 'chi_p']
low_level_coord_names=['mc','delta_mc','chi1_perp_bar','s1z_bar', 'phi1','chi2_perp_bar', 's2z_bar', 'phi2'] # assume this is the underlying

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

print("Cylinder test 1", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")


coord_names=['mc','delta_mc','xi','chiMinus','chi_p','s1x','s1y', 's2x', 's2y']
low_level_coord_names=['mc','delta_mc','chi1_perp_u','s1z_bar', 'phi1','chi2_perp_u', 's2z_bar', 'phi2'] # assume this is the underlying

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)
print(x2)

print("Cylinder test 2", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")


# Eccentric test
#   - assign variables
for P in P_list:
    P.eccentricity = np.random.uniform(0,0.3)
    P.meanPerAno = np.random.uniform(0, 2*np.pi)


coord_names=['eccentricity', 'meanPerAno', 'ecc_cos_meanPerAno', 'ecc_sin_meanPerAno']
low_level_coord_names=['eccentricity', 'meanPerAno'] # assume this is the underlying

x1 = np.zeros((npts,len(coord_names)))
x2 = np.zeros((npts,len(coord_names)))
y2 = np.zeros((npts,len(low_level_coord_names)))

for indx in np.arange(npts):
    P = P_list[indx]
    for indx_name  in np.arange(len(coord_names)):
        x1[indx,indx_name]  = P.extract_param( coord_names[indx_name])
    for indx_name2 in np.arange(len(low_level_coord_names)):
        y2[indx,indx_name2]  = P.extract_param( low_level_coord_names[indx_name2])

x2 = lalsimutils.convert_waveform_coordinates(y2, coord_names=coord_names, low_level_coord_names=low_level_coord_names)

print("Eccentric test ", np.max(np.abs(x1 - x2)))
err = np.max(np.abs(x1 - x2))
if opts.as_test and err > 1e-9:
    raise ValueError(" Large deviation seen ")



"""
test_precompute.py:  Testing the likelihood evaluation.
    - Generates a synthetic signal (at fixed or random sky location; at a fixed or random source orientation; at a fixed or random time)
      For simplicity, the source has fixed masses and distance.
    - The synthetic signal is introduced into a 3-detector network 'data' with *zero noise* and *an assumed known PSD*
    - Generates a 'template' signal at a reference distance
    - Precomputes all factors appearing in the likelihood
      Compares those factors for self-consistency (U and Q)
    - Evaluates the likelihood at the injection parameters and versus time

"""

__author__ = "Evan Ochsner <evano@gravity.phys.uwm.edu> and R. O'Shaughnessy"

import lalsimutils
import numpy as np
import lal
import lalsimulation as lalsim

import factored_likelihood
import factored_likelihood_test
from matplotlib import pylab as plt
import sys
import scipy.optimize

#
# Set by user
#
checkResults = True
checkResultsSlowChecks = True
checkInputPlots = False
checkResultsPlots = True
checkRhoIngredients = False
rosUseRandomSkyLocation = True
rosUseRandomSourceOrientation = True
rosUseRandomEventTime = False
rosUseRandomTemplateStartingFrequency = False

approxSignal = lalsim.TaylorT4
approxTemplate = lalsim.TaylorT4

#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
psd_dict = {}
rhoExpected ={}
rhoExpectedAlt ={}
analyticPSD_Q = True # For simplicity, using an analytic PSD

fminWaves = 25
fminWavesTemplate = 25
if (rosUseRandomTemplateStartingFrequency):
    print("   --- Generating a random template starting frequency  ---- ")
    fminWavesTemplate += 5*np.random.random_sample()
fminSNR = 25
fmaxSNR = 2000
fSample = 4096*4

theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 

psd_dict['H1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['L1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['V1'] = lalsim.SimNoisePSDiLIGOSRD


distanceFiducial = 25.  # Make same as reference
m1 = 4*lal.LAL_MSUN_SI
m2 = 3*lal.LAL_MSUN_SI
tEventFiducial = 0.0
if rosUseRandomEventTime:
    print("   --- Generating a random event (barycenter) time  ---- ")
    tEventFiducial+= 0.05*np.random.random_sample()
ampO =0 # sets which modes to include in the physical signal
Lmax = 2  # sets which modes to include in the output
fref = 100
    
Psig = lalsimutils.ChooseWaveformParams(fmin = fminWaves, radec=True, incl=0.0,phiref=0.0, theta=0.2, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         approx=approxSignal,
         fref=fref,
         tref=theEpochFiducial+tEventFiducial,
         deltaT=1./fSample,
        detector='H1', dist=distanceFiducial*1.e6*lal.LAL_PC_SI)
if rosUseRandomSkyLocation:
    print("   --- Generating a random sky location  ---- ")
    Psig.theta = np.arccos( 2*(np.random.random_sample())-1)
    Psig.phi = (np.random.random_sample())*2*lal.LAL_PI
    Psig.psi = (np.random.random_sample())*lal.LAL_PI
if rosUseRandomSourceOrientation:
    print("   --- Generating a random source orientation  ---- ")
    Psig.incl = np.arccos( 2*(np.random.random_sample())-1)
    Psig.phiref = (np.random.random_sample())*2*lal.LAL_PI
if rosUseRandomEventTime:
    print("   --- Generating a random event (barycenter) time  ---- ")
    Psig.tref += np.random.random_sample()

df = lalsimutils.findDeltaF(Psig)
Psig.deltaF = df
Psig.print_params()
print(" ======= Generating synthetic data in each interferometer (manual timeshifts) ==========")
t0 = Psig.tref
Psig.detector = 'H1'
data_dict['H1'] = lalsimutils.non_herm_hoff(Psig)  # already takes care of propagating to a detector, using the 'detector' field
Psig.detector = 'L1'
data_dict['L1'] = lalsimutils.non_herm_hoff(Psig)
Psig.detector = 'V1'
data_dict['V1'] = lalsimutils.non_herm_hoff(Psig)


print(" == Data report == ")
detectors = data_dict.keys()
rho2Net = 0
for det in detectors:
    IP = lalsimutils.ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det])
    IPOverlap = lalsimutils.ComplexOverlap(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det],analyticPSD_Q=True,full_output=True)  # Use for debugging later
    rhoExpected[det] = rhoDet = IP.norm(data_dict[det])
    rhoExpectedAlt[det] = rhoDet2 = IPOverlap.norm(data_dict[det])
    rho2Net += rhoDet*rhoDet
    print(det, rhoDet, rhoDet2, " has arrival time relative to fiducial of ", float(factored_likelihood.ComputeArrivalTimeAtDetector(det, Psig.phi,Psig.theta,Psig.tref) - theEpochFiducial))
    tarrive = factored_likelihood.ComputeArrivalTimeAtDetector(det, Psig.phi,Psig.theta,Psig.tref)
    print(" and has  epoch ", float(data_dict[det].epoch), " with arrival time ",  lalsimutils.stringGPSNice(tarrive))
print("Network : ", np.sqrt(rho2Net))

if checkInputPlots:
    print(" == Plotting detector data (time domain; requires regeneration, MANUAL TIMESHIFTS,  and seperate code path! Argh!) == ")
    P = Psig.copy()
    P.tref = Psig.tref
    for det in detectors:
        P.detector=det   # we do 
        hT = hoft(P)
        tvals = float(P.tref - theEpochFiducial) + hT.deltaT*np.arange(len(hT.data.data))
        plt.figure(1)
        plt.plot(tvals, hT.data.data,label=det)

    tlen = hT.deltaT*len(np.nonzero(np.abs(hT.data.data)))
    plt.xlim( 9.5,11)  # Not well centered
    plt.show()



print(" ======= Template specified: precomputing all quantities ==========")
# Struct to hold template parameters
# Fiducial distance provided but will not be used
P =  lalsimutils.ChooseWaveformParams(fmin=fminWavesTemplate, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         approx=approxTemplate,
         fref=fref,
         deltaT=1./fSample,
         tref=theEpochFiducial,
         deltaF=df)
rholms_intp, crossTerms, rholms, epoch_post = factored_likelihood.PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict, psd_dict, Lmax, fmaxSNR, analyticPSD_Q)


TestDictionary = factored_likelihood_test.TestDictionaryDefault
TestDictionary["DataReport"]             = True
TestDictionary["DataReportTime"]       = False
TestDictionary["UVReport"]              = True
TestDictionary["UVReflection"]          = True
TestDictionary["QReflection"]          = False
TestDictionary["lnLModelAtKnown"]  = True
TestDictionary["lnLDataAtKnownPlusOptimalTimePhase"] = False
TestDictionary["lnLAtKnown"]           = True
TestDictionary["lnLAtKnownMarginalizeTime"]  = False
TestDictionary["lnLDataPlot"]            = True

#opts.fmin_SNR=40

factored_likelihood_test.TestLogLikelihoodInfrastructure(TestDictionary,theEpochFiducial, epoch_post, data_dict, psd_dict, fmaxSNR, analyticPSD_Q, Psig, rholms,rholms_intp, crossTerms, detectors,Lmax)
sys.exit(0)




"""
test_like_and_samp_noisydata.py:  Testing convergnece of intrinsic marginalization with noisy, 3-ifo data.
  - Loads frame data that Chris generated (assumed in signal_hoft) into 3-detector data sets
  - Uses an assumed PSD
  - Constructs a template which *exactly* matches the known source
  - 
"""

from __future__ import print_function

from pylal import Fr
import sys
from optparse import OptionParser

import numpy as np
from matplotlib import pylab as plt

from glue.lal import Cache
import lal
import lalsimulation as lalsim
import lalsimutils as lsu


__author__ = "Evan Ochsner <evano@gravity.phys.uwm.edu>, Chris Pankow <pankow@gravity.phys.uwm.edu>, R. O'Shaughnessy"

import factored_likelihood
import ourio

checkInputs = True
checkResults = True

rosUseZeroNoiseCache  = False


rosUseDifferentWaveformLengths = False    
rosUseRandomTemplateStartingFrequency = False

rosUseTargetedDistance = True
rosUseStrongPriorOnParameters = False
rosDebugMessages = False
rosShowSamplerInputDistributions = True
rosShowRunningConvergencePlots = True
rosShowTerminalSampleHistograms = True
rosUseThresholdForReturn = False
rosSaveHighLikelihoodPoints = True
nMaxEvals = 1e4

if rosUseThresholdForReturn and nMaxEvals > 5e4:  # I usually only want to restrict the return set when I am testing
    fracThreshold = 0.95
else:
    fracThreshold = 0.0


theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 
tEventFiducial = 0                                                                 # relative to GPS reference

optp = OptionParser()
optp.add_option("-c", "--cache-file", default=None, help="LIGO cache file containing all data needed.")
optp.add_option("-C", "--channel-name", action="append", help="instrument=channel-name, e.g. H1=FAKE-STRAIN. Can be given multiple times for different instruments.")
opts, args = optp.parse_args()

data_dict = {}
data_dict_time = {}
psd_dict = {}
rhoExpected ={}
rhoManual ={}
rhoExpectedAlt ={}
analyticPSD_Q = True # For simplicity, using an analytic PSD
# Note: Injected signal has fmin = 25. We should choose a different value.
fminWaves = 30
fminSNR = 30
fSample = 4096*4  # will be reset by data sampling rate.


Niter = 5 # Number of times to call likelihood function
Tmax = 0.05 # max ref. time
Tmin = -0.05 # min ref. time
Dmax = 110. * 1.e6 * lal.LAL_PC_SI # max distance
Dmin = 1. * 1.e6 * lal.LAL_PC_SI   # min distance

ampO =0 # sets which modes to include in the physical signal
Lmax = 2 # sets which modes to include
fref = 100
fminWavesSignal = 25  # too long can be a memory and time hog, particularly at 16 kHz

if rosUseDifferentWaveformLengths: 
    fminWavesTemplate = fminWavesSignal+0.005
else:
    if rosUseRandomTemplateStartingFrequency:
         print("   --- Generating a random template starting frequency  ---- ")
         fminWavesTemplate = fminWavesSignal+5.*np.random.random_sample()
    else:
        fminWavesTemplate = fminWavesSignal


theEpochFiducial = lal.LIGOTimeGPS(1000000014.000000000)   # Use actual injection GPS time (assumed from trigger)
#theEpochFiducial = lal.LIGOTimeGPS(1000000000.000000000)     # Use epoch of the data
tEventFiducial = 0   #  time relative to fiducial epoch, used to identify window to look in.  Checked empirically.

detectors = ['H1', "L1", "V1"]
psd_dict['H1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['L1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['V1'] = lalsim.SimNoisePSDiLIGOSRD

# Load IFO FFTs.
# ASSUME data has same sampling rate!
if not rosUseZeroNoiseCache:
    fnameCache = 'test1.cache'
else:
    fnameCache = 'test1-0noise.cache'
data_dict['H1']=lsu.frame_data_to_non_herm_hoff(fnameCache, "H1"+":FAKE-STRAIN")
data_dict['V1']=lsu.frame_data_to_non_herm_hoff(fnameCache, "V1"+":FAKE-STRAIN")
data_dict['L1']=lsu.frame_data_to_non_herm_hoff(fnameCache, "L1"+":FAKE-STRAIN")
#print data_dict['H1'].data.data[10]  # confirm data loaded
df = data_dict['H1'].deltaF  
fSample = len(data_dict['H1'].data.data)*data_dict['H1'].deltaF  # Note two-sided
print(" sampling rate of data = ", fSample)
print(" ===  Repeating SNR calculation (only valid for noiseless data) ===")
rho2NetManual = 0
for det in detectors:
    IP = lsu.ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det])
    rhoManual[det] = IP.norm(data_dict[det])
    rho2NetManual+= rhoManual[det]*rhoManual[det]
    print(det, rhoManual[det], " via raw data")


# TARGET INJECTED SIGNAL (for reference and calibration of results)
Psig = lsu.xml_to_ChooseWaveformParams_array("mdc.xml.gz")[0]  # Load in the physical parameters of the injection (=first element)
dfSig = Psig.deltaF = lsu.findDeltaF(Psig)  # NOT the full deltaf to avoid padding problems
Psig.print_params()
print(" == Data report == ")
rho2Net = 0
for det in detectors:
    Psig.detector = det
    hT = lsu.non_herm_hoff(Psig)
    fSampleSig = len(hT.data.data)*hT.deltaF
    IP = lsu.ComplexIP(fLow=fminSNR, fNyq=fSampleSig/2,deltaF=dfSig,psd=psd_dict[det])
    rhoExpected[det] = rhoDet = IP.norm(hT)
    rho2Net += rhoDet*rhoDet
    print(det, rhoDet, "compare to result from raw data: ", rhoManual[det],  "; note it arrival time relative to fiducial of ", float(factored_likelihood.ComputeArrivalTimeAtDetector(det, Psig.phi,Psig.theta,Psig.tref) - theEpochFiducial))
print("Network : ", np.sqrt(rho2Net))


if checkInputs:
    print(" == Plotting detector data (time domain; requires regeneration, MANUAL TIMESHIFTS,  and seperate code path! Argh!) == ")
    P = Psig.copy()
    P.tref = Psig.tref
    for det in detectors:
        P.detector=det   # we do 
        hT = lsu.hoft(P)
        tvals = float(P.tref - theEpochFiducial) + hT.deltaT*np.arange(len(hT.data.data))
        plt.figure(1)
        plt.plot(tvals, hT.data.data,label=det)

    tlen = hT.deltaT*len(np.nonzero(np.abs(hT.data.data)))
    tRef = np.abs(float(hT.epoch))
#    plt.xlim( tRef-0.5,tRef+0.1)  # Not well centered, based on epoch to identify physical merger time
    plt.savefig("test_like_and_samp-input-hoft.pdf")


m1 = 4*lal.LAL_MSUN_SI
m2 = 4*lal.LAL_MSUN_SI
ampO =0 # sets which modes to include in the physical signal
Lmax = 2  # sets which modes to include in the output
fref = 100

# Struct to hold template parameters
P = lsu.ChooseWaveformParams(fmin=fminWavesTemplate, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         fref=fref,
         tref=theEpochFiducial,
         deltaT=1./fSample,
         dist=100*1.e6*lal.LAL_PC_SI,
         deltaF=df)


#
# Perform the Precompute stage
#
rholms_intp, crossTerms, rholms, epoch_post = factored_likelihood.PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict,psd_dict, Lmax, analyticPSD_Q)
print("Finished Precomputation...")
print("====Generating metadata from precomputed results =====")
distBoundGuess = factored_likelihood.estimateUpperDistanceBoundInMpc(rholms, crossTerms)
print(" distance probably less than ", distBoundGuess, " Mpc")

print("====Loading metadata from previous runs (if any): sampler-seed-data.dat =====")


if checkInputs == True:

    print(" ======= UV test: Recover the SNR of the injection  ==========")
    print(" Detector lnLmodel  (-2lnLmodel)^(1/2)  rho(directly)  [last two entries should be equal!] ")
    for det in detectors:
        lnLModel = factored_likelihood.SingleDetectorLogLikelihoodModel(crossTerms, Psig.tref, Psig.phi, Psig.theta, Psig.incl, Psig.phiref, Psig.psi, Psig.dist, 2, det)
        print(det, lnLModel, np.sqrt(-2*lnLModel), rhoExpected[det], "      [last two equal?]")
    print(" ======= End to end LogL: Recover the SNR of the injection at the injection parameters  ==========")
    lnL = factored_likelihood.FactoredLogLikelihood(Psig, rholms_intp, crossTerms, Lmax)
    print("  : Evan's code : ", lnL, " versus rho^2/2 ", rho2Net/2)
    print("  : Timing issues (checkme!) : fiducial = ", lsu.stringGPSNice(theEpochFiducial))

    print(" ======= rholm test: Plot the lnLdata timeseries at the injection parameters (* STILL TIME OFFSET *)  ==========")
    tmin = np.max(float(epoch_post - theEpochFiducial),factored_likelihood.tWindowReference[0]+0.03)   # the minimum time used is set by the rolling condition
#    tvals = np.linspace(tmin,factored_likelihood.tWindowReference[1],4*fSample*(factored_likelihood.tWindowReference[1]-tmin))
    tvals = np.linspace(factored_likelihood.tWindowExplore[0]+tEventFiducial,factored_likelihood.tWindowExplore[1]+tEventFiducial,fSample*(factored_likelihood.tWindowExplore[1]-factored_likelihood.tWindowExplore[0]))
    for det in detectors:
        lnLData = map( lambda x: factored_likelihood.SingleDetectorLogLikelihoodData(theEpochFiducial,rholms_intp, theEpochFiducial+x, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, 2, det), tvals)
        lnLDataEstimate = np.ones(len(tvals))*rhoExpected[det]*rhoExpected[det]
        plt.figure(1)
        tvalsPlot = tvals 
        plt.plot(tvalsPlot, lnLData,label='Ldata(t)+'+det)
        plt.plot(tvalsPlot, lnLDataEstimate,label="$rho^2("+det+")$")
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is ", tEventRelative)
    plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
    plt.title("lnLdata (interpolated) vs narrow time interval")
    plt.xlabel('t(s)')
    plt.ylabel('lnLdata')

    print(" ======= rholm test: Plot the lnL timeseries at the injection parameters (* STILL TIME)  ==========")
    tmin = np.max(float(epoch_post - theEpochFiducial),factored_likelihood.tWindowReference[0]+0.03)   # the minimum time used is set by the rolling condition
    tvals = np.linspace(tmin,factored_likelihood.tWindowReference[1],4*fSample*(factored_likelihood.tWindowReference[1]-tmin))
    P = Psig.copy()
    lnL = np.zeros(len(tvals))
    for indx in np.arange(len(tvals)):
            P.tref =  theEpochFiducial+tvals[indx]
            lnL[indx] =  factored_likelihood.FactoredLogLikelihood(P, rholms_intp, crossTerms, 2)
    lnLEstimate = np.ones(len(tvals))*rho2Net/2
    plt.figure(1)
    tvalsPlot = tvals 
    plt.plot(tvalsPlot, lnL,label='lnL(t)')
    plt.plot(tvalsPlot, lnLEstimate,label="$rho^2/2(net)$")
    tEventRelative =float( Psig.tref - theEpochFiducial)
    print(" Real time (relative to fiducial start time) ", tEventFiducial,  " and our triggering time is the same ", tEventRelative)
    plt.plot([tEventFiducial,tEventFiducial],[0,rho2Net], color='k',linestyle='--')
    plt.title("lnL (interpolated) vs narrow time interval")
    plt.xlim(Tmin,Tmax)  # the window we actually use
    plt.legend()
    plt.savefig("test_like_and_samp-lnLvsTime.pdf")


#
# Call the likelihood function for various extrinsic parameter values
#
nEvals = 0
def likelihood_function(phi, theta, tref, phiref, incl, psi, dist):
    global nEvals
    global pdfFullPrior

    lnL = np.zeros(phi.shape)
    i = 0
#    print " Likelihood results :  "
#    print " iteration Neff  lnL   sqrt(2max(lnL))  rho  sqrt(2 lnLmarg)   <lnL> "
    for ph, th, tr, phr, ic, ps, di in zip(phi, theta, tref, phiref, incl, psi, dist):
        P.phi = ph # right ascension
        P.theta = th # declination
        P.tref = theEpochFiducial + tr # ref. time (rel to epoch for data taking)
        P.phiref = phr # ref. orbital phase
        P.incl = ic # inclination
        P.psi = ps # polarization angle
        P.dist = di # luminosity distance

        lnL[i] = factored_likelihood.FactoredLogLikelihood(P, rholms_intp, crossTerms, Lmax)#+ np.log(pdfFullPrior(ph, th, tr, ps, ic, ps, di))
#        if i<len(phi)-10:
#            LSum[i+1] = LSum[i]+np.exp(lnL[i])
#        if (np.mod(i,1000)==10 and i>100):
#            print " iteration Neff  lnL   sqrt(2max(lnL)) rho  sqrt(2 lnLmarg)   <lnL> "  # reminder
        # if (np.mod(i,200)==10 and i>100):
        #     Neff = LSum[i+1]/np.exp(np.max(lnL[:i]))   # should actually include sampling distribution and prior distribution correction in it
            # if rosDebugMessages:
            #     print "\t Params ", nEvals+i, " (RA, DEC, tref, phiref, incl, psi, dist) ="
            #     print "\t", i, P.phi, P.theta, float(P.tref-theEpochFiducial), P.phiref, P.incl, P.psi, P.dist/(1e6*lal.LAL_PC_SI), lnL[i]
            #     print "\t sampler probability of draws", sampler.cdf['ra'](P.phi), sampler.cdf['dec'](P.theta),sampler.cdf['tref'](float(P.tref - theEpochFiducial)), sampler.cdf['phi'](P.phiref), sampler.cdf['incl'](P.incl), sampler.cdf['psi'](P.psi), sampler.cdf['dist'](P.dist)
            # logLmarg =np.log(np.mean(np.exp(lnL[:i])))
#            print nEvals+i, Neff,  lnL[i],   np.sqrt(2*np.max(lnL[:i])), np.sqrt(rho2Net), np.sqrt(2*logLmarg), np.mean(lnL[:i])
        i+=1


    nEvals+=i 
    return np.exp(lnL)

import RIFT.integrators.mcsampler as mcsampler
sampler = mcsampler.MCSampler()

# Sampling distribution
def uniform_samp(a, b, x):
   if type(x) is float:
       return 1/(b-a)
   else:
       return np.ones(x.shape[0])/(b-a)

# set up bounds on parameters
# Polarization angle
psi_min, psi_max = 0, np.pi
# RA and dec
ra_min, ra_max = 0, 2*np.pi
dec_min, dec_max = -np.pi/2, np.pi/2
# Reference time
tref_min, tref_max = Tmin, Tmax
# Inclination angle
inc_min, inc_max = 0, np.pi
# orbital phi
phi_min, phi_max = 0, 2*np.pi
# distance
dist_min, dist_max = Dmin, Dmax

import functools
# Define the true prior PDF
# The likelihood function assumes this function exists.
# The integrator actually does a monte carlo integral of L*p using a sampling prior p_s.
def pdfFullPrior(phi, theta, tref, phiref, incl, psi, dist): # remember theta is dec, 
    return (np.cos(theta)/(4*np.pi)) * 1./(factored_likelihood.tWindowExplore[1]-factored_likelihood.tWindowExplore[0])*1/(3*Dmax**3)*np.cos(incl)/(4*np.pi)*1./(np.pi)


# Uniform sampling (in area) but nonuniform sampling in distance (*I hope*).  Auto-cdf inverse
if rosUseStrongPriorOnParameters:
    sampler.add_parameter("psi", functools.partial(mcsampler.gauss_samp, Psig.psi, 0.5), None, psi_min, psi_max, 
                          prior_pdf =mcsampler.uniform_samp_psi  )
    # Use few degree square prior : at SNR 7 in 3 detetors.  This is about 10 deg square*4
    sampler.add_parameter("ra", functools.partial(mcsampler.gauss_samp, Psig.phi,0.05*2), None, ra_min, ra_max, 
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("dec", functools.partial(mcsampler.gauss_samp, Psig.theta,0.05*2), None, dec_min, dec_max, 
                          prior_pdf= mcsampler.uniform_samp_dec)
    sampler.add_parameter("tref", functools.partial(mcsampler.gauss_samp, tEventFiducial, 0.005*2), None, tref_min, tref_max, 
                          prior_pdf = functools.partial(mcsampler.uniform_samp_vector, factored_likelihood.tWindowExplore[0],factored_likelihood.tWindowExplore[1]))
    sampler.add_parameter("phi", functools.partial(mcsampler.gauss_samp, Psig.phiref,0.5), None, phi_min, phi_max, 
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("incl", functools.partial(mcsampler.gauss_samp, Psig.incl,0.3), None, inc_min, inc_max, 
                          prior_pdf = mcsampler.uniform_samp_theta)
    if rosUseTargetedDistance:
        sampler.add_parameter("dist", functools.partial(mcsampler.quadratic_samp_vector,  distBoundGuess*1e6*lal.LAL_PC_SI ), None, dist_min, dist_max, prior_pdf = np.vectorize(lambda x: x**2/(3*Dmax**3)))
    else:
        sampler.add_parameter("dist", functools.partial(mcsampler.gauss_samp_withfloor, Psig.dist, Psig.dist*0.1, 0.001/Psig.dist), None, dist_min, dist_max,prior_pdf =np.vectorize( lambda x: x**2/(3*Dmax**3)))
else:
    # PROBLEM: Underlying prior samplers are not uniform.  We need two stages
    sampler.add_parameter("psi", functools.partial(mcsampler.uniform_samp_vector, psi_min, psi_max), None, psi_min, psi_max,
                          prior_pdf =mcsampler.uniform_samp_psi )
    # Use few degree square prior : at SNR 20 in 3 detetors.  This is about 10 deg square
    sampler.add_parameter("ra", mcsampler.uniform_samp_phase, None, ra_min, ra_max, 
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("dec", mcsampler.uniform_samp_dec, None, dec_min, dec_max, 
                          prior_pdf= mcsampler.uniform_samp_dec)
    sampler.add_parameter("tref", functools.partial(mcsampler.gauss_samp, tEventFiducial, 0.005*2), None, tref_min, tref_max, 
                          prior_pdf = functools.partial(mcsampler.uniform_samp_vector, factored_likelihood.tWindowExplore[0],factored_likelihood.tWindowExplore[1]))
    sampler.add_parameter("phi", functools.partial(mcsampler.uniform_samp_vector, phi_min, phi_max), None, phi_min, phi_max,
                          prior_pdf = mcsampler.uniform_samp_phase)
    sampler.add_parameter("incl", functools.partial(mcsampler.cos_samp_vector), None, inc_min, inc_max,
                          prior_pdf = mcsampler.uniform_samp_theta)
    if rosUseTargetedDistance:
        sampler.add_parameter("dist", functools.partial(mcsampler.quadratic_samp_vector,  distBoundGuess*1e6*lal.LAL_PC_SI ), None, dist_min, dist_max, prior_pdf = np.vectorize(lambda x: x**2/(3*Dmax**3)))
    else:
        sampler.add_parameter("dist", functools.partial(mcsampler.gauss_samp_withfloor, Psig.dist, Psig.dist*0.1, 0.001/Psig.dist), None, dist_min, dist_max,prior_pdf =np.vectorize( lambda x: x**2/(3*Dmax**3)))


if rosShowSamplerInputDistributions:
    print(" ====== Plotting prior and sampling distributions ==== ")
    print("  PROBLEM: Build in/hardcoded via uniform limits on each parameter! Need to add measure factors ")
    nFig = 0
    for param in sampler.params:
        nFig+=1
        plt.figure(nFig)
        plt.clf()
        xLow = sampler.llim[param]
        xHigh = sampler.rlim[param]
        xvals = np.linspace(xLow,xHigh,500)
        pdfPrior = sampler.prior_pdf[param]  # Force type conversion in case we have non-float limits for some reasona
        pdfvalsPrior = np.array(map(pdfPrior, xvals))  # do all the np operations by hand: no vectorization
        pdf = sampler.pdf[param]
        cdf = sampler.cdf[param]
        pdfvals = pdf(xvals)
        cdfvals = cdf(xvals)
        if str(param) ==  "dist":
            xvvals = xvals/(1e6*lal.LAL_PC_SI)       # plot in Mpc, not m.  Note PDF has to change
            pdfvalsPrior = pdfvalsPrior * (1e6*lal.LAL_PC_SI) # rescale units
            pdfvals = pdfvals * (1e6*lal.LAL_PC_SI) # rescale units
        plt.plot(xvals,pdfvalsPrior,label="prior:"+str(param),linestyle='--')
        plt.plot(xvals,pdfvals,label=str(param))
        plt.plot(xvals,cdfvals,label='cdf:'+str(param))
        plt.xlabel(str(param))
        plt.legend()
        plt.savefig("test_like_and_samp-"+str(param)+".pdf")
#    plt.show()

tGPSStart = lal.GPSTimeNow()
res, var, ret, lnLmarg, neff = sampler.integrate(likelihood_function, "ra", "dec", "tref", "phi", "incl", "psi", "dist", n=100,nmax=nMaxEvals,igrandmax=rho2Net/2,full_output=True,neff=100,igrand_threshold_fraction=fracThreshold,verbose=True)
tGPSEnd = lal.GPSTimeNow()
print(" Evaluation time  = ", float(tGPSEnd - tGPSStart), " seconds")
print(" lnLmarg is ", np.log(res), " with expected relative error ", np.sqrt(var)/res)
print(" expected largest value is ", rho2Net/2, end=' ')
print(" note neff is ", neff, "; compare neff^(-1/2) = ", 1/np.sqrt(neff))

# Save the sampled points to a file
# Only store some
ourio.dumpSamplesToFile("test_like_and_samp_noisydata-dump.dat", ret, ['ra','dec', 'tref', 'phi', 'incl', 'psi', 'dist', 'lnL'])

if checkInputs:
    plt.show()
# Plot terminal histograms from the sampled points and log likelihoods
if rosShowTerminalSampleHistograms:
    ra,dec,tref,phi,incl, psi,dist,lnL = np.transpose(ret)  # unpack. This can include all or some of the data set. The default configuration returns *all* points
    print(" ==== CONVERGENCE PLOTS (**beware: potentially truncated data!**) === ")
    plt.figure(0)
    plt.clf()
    plt.plot(np.arange(len(lnLmarg)), lnLmarg,label="lnLmarg")
    plt.plot(np.arange(len(lnL)), np.ones(len(lnL))*rho2Net/2,label="rho^2/2")
    plt.xlim(0,len(lnL))
    plt.xlabel('iteration')
    plt.ylabel('lnL')
    plt.legend()
    ourio.plotParameterDistributionsFromSamples("results", sampler, ret,  ['ra','dec', 'tref', 'phi', 'incl', 'psi', 'dist', 'lnL'])
    # print " ==== TERMINAL 1D HISTOGRAMS: Sampling and posterior === "
    # plt.figure(1)
    # plt.clf()
    # hist, bins  = np.histogram(dist/(1e6*lal.LAL_PC_SI),bins=50,density=True)
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="dist:sampled")
    # hist,bins = np.histogram(dist/(1e6*lal.LAL_PC_SI),bins=50,weights=np.exp(lnL),density=True)
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="dist:post")
    # plt.xlabel("d (Mpc)")
    # plt.title("Sampling and posterior distribution: d ")
    # plt.legend()
    # plt.figure(2)
    # plt.clf()
    # hist, bins  = np.histogram(tref,bins=50,density=True)
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="tref:sampled")
    # hist, bins  = np.histogram(tref,bins=50,density=True,weights=np.exp(lnL))
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="tref:post")
    # plt.xlim(-0.01+tEventFiducial,0.01+tEventFiducial)
    # plt.xlabel("t (s)")
    # plt.title("Sampling and posterior distribution: t ")
    # plt.legend()
    # plt.figure(3)
    # plt.clf()
    # hist, bins  = np.histogram(incl,bins=50,normed=True)
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="incl:sampled")
    # hist, bins  = np.histogram(incl,bins=50,normed=True,weights=np.exp(lnL))
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="incl:post")
    # plt.xlabel("incl")
    # plt.title("Sampling and posterior distribution: incl ")
    # plt.legend()
    # plt.figure(4)
    # plt.clf()
    # hist, bins  = np.histogram(psi,bins=50,normed=True)
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="psi:sampled")
    # hist, bins  = np.histogram(psi,bins=50,normed=True,weights=np.exp(lnL))
    # center = (bins[:-1]+bins[1:])/2
    # plt.plot(center,hist,label="psi:post")
    # plt.xlabel("psi")
    # plt.title("Sampling and posterior distribution: psi ")
    # plt.legend()
    # plt.show()
    print(" ==== TERMINAL 2D HISTOGRAMS: Sampling and posterior === ")
        # Distance-inclination
    plt.figure(1)
    plt.clf()
    H, xedges, yedges = np.histogram2d(dist/(1e6*lal.LAL_PC_SI),incl, weights=np.exp(lnL),bins=(10,10))
    extent = [yedges[0], yedges[-1], xedges[-1], xedges[0]]
    plt.imshow(H, extent=extent, interpolation='nearest', aspect=0.618)
    plt.colorbar()
    plt.xlim(0, 50)
    plt.ylim(0, np.pi)
    plt.title("Posterior distribution: d-incl ")
        # phi-psi
    plt.figure(2)
    plt.clf()
    H, xedges, yedges = np.histogram2d(phi,psi, bins=(10,10),weights=np.exp(lnL))
    extent = [yedges[0], yedges[-1], xedges[-1], xedges[0]]
    plt.imshow(H, extent=extent, interpolation='nearest', aspect=0.618)
    plt.colorbar()
    plt.xlim(0,2*np.pi)
    plt.ylim(0,np.pi)
    plt.title("Posterior distribution: phi-psi ")
        # ra-dec
    plt.figure(3)
    plt.clf()
    H, xedges, yedges = np.histogram2d(ra,dec, bins=(10,10),weights=np.exp(lnL))
    extent = [yedges[0], yedges[-1], xedges[-1], xedges[0]]
    plt.imshow(H, extent=extent, interpolation='nearest', aspect=0.618)
    plt.xlim(0,2*np.pi)
    plt.ylim(-np.pi,np.pi)
    plt.colorbar()
    plt.title("Posterior distribution: ra-dec ")
    plt.show()



# From https://github.com/jellis18/PAL/blob/master/bayesutils.py
#   - modified to minimize non-lalsuite installations
#   - requires
#        -healpy
#        -statsmodels http://statsmodels.sourceforge.net/, which requires pandas
import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as interp
import scipy.ndimage.filters as filter
import healpy as hp
from lalinference.bayestar import plot as bplot
import matplotlib.mlab as ml
#import statsmodels.api as sm
from matplotlib.ticker import FormatStrFormatter, LinearLocator, NullFormatter, NullLocator
import matplotlib.ticker
import matplotlib.colors
from optparse import OptionParser

"""
Given a 2D matrix of (marginalised) likelihood levels, this function returns
the 1, 2, 3- sigma levels. The 2D matrix is usually either a 2D histogram or a
likelihood scan

"""
def getsigmalevels(hist2d):
  # We will draw contours with these levels
  sigma1 = 0.68268949
  level1 = 0
  sigma2 = 0.95449974
  level2 = 0
  sigma3 = 0.99730024
  level3 = 0

  #
  lik = hist2d.reshape(hist2d.size)
  sortlik = np.sort(lik)

  # Figure out the 1sigma level
  dTotal = np.sum(sortlik)
  nIndex = sortlik.size
  dSum = 0
  while (dSum < dTotal * sigma1):
    nIndex -= 1
    dSum += sortlik[nIndex]
  level1 = sortlik[nIndex]

  # 2 sigma level
  nIndex = sortlik.size
  dSum = 0
  while (dSum < dTotal * sigma2):
    nIndex -= 1
    dSum += sortlik[nIndex]
  level2 = sortlik[nIndex]

  # 3 sigma level
  nIndex = sortlik.size
  dSum = 0
  while (dSum < dTotal * sigma3):
    nIndex -= 1
    dSum += sortlik[nIndex]
  level3 = sortlik[nIndex]

  return level1, level2, level3

# def confinterval(samples, sigma=0.68, onesided=False):
#     """

#     Given a list of samples, return the desired cofidence intervals.
#     Returns the minimum and maximum confidence levels

#     @param samples: Samples that we wish to get confidence intervals

#     @param sigmalevel: Sigma level 1, 2, or 3 sigma, will return 
#                        corresponding confidence limits

#     @param onesided: Boolean to use onesided or twosided confidence
#                      limits.

#     """

#     # Create the ecdf function
#     ecdf = sm.distributions.ECDF(samples)

#     # Create the binning
#     x = np.linspace(min(samples), max(samples), 1000)
#     y = ecdf(x)

#     # Find the intervals
#     x2min = y[0]
#     if onesided:
#         bound = 1 - sigma
#     else:
#         bound = 0.5*(1-sigma)

#     for i in range(len(y)):
#         if y[i] >= bound:
#             x2min = x[i]
#             break

#     x2max = y[-1]
#     if onesided:
#         bound = sigma
#     else:
#         bound = 1 - 0.5 * (1 - sigma)

#     for i in reversed(range(len(y))):
#         if y[i] <= bound:
#             x2max = x[i]
#             break

#     return x2min, x2max



def makesubplot2d(ax, samples1, samples2, color=True, weights=None, smooth=True, \
                  bins=[40, 40], contours=True, x_range=None, y_range=None, \
                  logx=False, logy=False, logz=False):
    
    if x_range is None:
        xmin = np.min(samples1)
        xmax = np.max(samples1)
    else:
        xmin = x_range[0]
        xmax = x_range[1]

    if y_range is None:
        ymin = np.min(samples2)
        ymax = np.max(samples2)
    else:
        ymin = y_range[0]
        ymax = y_range[1]

    if logx:
        bins[0] = np.logspace(np.log10(xmin), np.log10(xmax), bins[0])
    
    if logy:
        bins[1] = np.logspace(np.log10(ymin), np.log10(ymax), bins[1])

    hist2d,xedges,yedges = np.histogram2d(samples1, samples2, weights=weights, \
            bins=bins,range=[[xmin,xmax],[ymin,ymax]])
    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1] ]

    if logz:
        for ii in range(hist2d.shape[0]):
            for jj in range(hist2d.shape[1]):
                if hist2d[ii,jj] <= 0:
                    hist2d[ii,jj] = 1

    
    xedges = np.delete(xedges, -1) + 0.5*(xedges[1] - xedges[0])
    yedges = np.delete(yedges, -1) + 0.5*(yedges[1] - yedges[0])
    
    # gaussian smoothing
    if smooth:
        hist2d = filter.gaussian_filter(hist2d, sigma=0.75)

    if contours:
        
        level1, level2, level3 = getsigmalevels(hist2d)
        
        contourlevels = (level1, level2, level3)
        
        #contourcolors = ('darkblue', 'darkblue', 'darkblue')
        contourcolors = ('black', 'black', 'black')
        contourlinestyles = ('-', '--', ':')
        contourlinewidths = (1.5, 1.5, 1.5)
        contourlabels = [r'1 $\sigma$', r'2 $\sigma$',r'3 $\sigma$']
        
        contlabels = (contourlabels[0], contourlabels[1], contourlabels[2])

        c1 = ax.contour(xedges,yedges,hist2d.T,contourlevels, \
                colors=contourcolors, linestyles=contourlinestyles, \
                linewidths=contourlinewidths, zorder=2)
    if color:
        if logz:
            c2 = ax.imshow(np.flipud(hist2d.T), extent=extent, aspect=ax.get_aspect(), \
                      interpolation='gaussian', norm=matplotlib.colors.LogNorm())
        else:
            c2 = ax.imshow(np.flipud(hist2d.T), extent=extent, aspect=ax.get_aspect(), \
                      interpolation='gaussian')

    if logx:
        ax.set_xscale('log')
    if logy:
        ax.set_yscale('log')
    
    
def makesubplot1d(ax, samples, weights=None, interpolate=False, smooth=True,\
                  label=None, bins=30, range=None, color='k'):
    """ 
    Make histogram of samples

    """

    if range is None:
        hist, xedges = np.histogram(samples, bins, normed=True, weights=weights)
    else:
        hist, xedges = np.histogram(samples, bins, normed=True, range=range, weights=weights)

    xedges = np.delete(xedges, -1) + 0.5*(xedges[1] - xedges[0])

    # gaussian smoothing
    if smooth:
        hist = filter.gaussian_filter(hist, sigma=0.75)
        if interpolate:
            f = interp.interp1d(xedges, hist, kind='cubic')
            xedges = np.linspace(xedges.min(), xedges.max(), 10000)
            hist = f(xedges)

    # make plot
    if label is not None:
        ax.plot(xedges, hist, color=color, lw=1.5, label=label)
    else:
        ax.plot(xedges, hist, color=color, lw=1.5)

def getMax(samples, weights=None, range=None, bins=50):
    """ 
    Make histogram of samples

    """

    if range is None:
        hist, xedges = np.histogram(samples, bins, normed=True)
    else:
        hist, xedges = np.histogram(samples, bins, normed=True, range=range)

    xedges = np.delete(xedges, -1) + 0.5*(xedges[1] - xedges[0])

    # gaussian smoothing
    hist = filter.gaussian_filter(hist, sigma=0.75)

    # interpolation
    f = interp.interp1d(xedges, hist, kind='cubic')
    xedges = np.linspace(xedges.min(), xedges.max(), 10000)
    hist = f(xedges)

    return xedges[np.argmax(hist)]
        

# make triangle plot of marginalized posterior distribution
def triplot(chain, color=True, weights=None, interpolate=False, smooth=True, \
           labels=None, figsize=(11,8.5), title=None, inj=None):

    """

    Make Triangle plot

    """

    # rcParams settings
    plt.rcParams['ytick.labelsize'] = 10.0
    plt.rcParams['xtick.labelsize'] = 10.0
    plt.rcParams['text.usetex'] = True
    plt.rcParams['figure.figsize'] = figsize

    # get number of parameters
    ndim = chain.shape[1]
    parameters = np.linspace(0,ndim-1,ndim)

    f, axarr = plt.subplots(nrows=len(parameters), ncols=len(parameters),figsize=figsize)

    for i in range(len(parameters)):
        # for j in len(parameters[np.where(i <= parameters)]:
        for j in range(len(parameters)):
            ii = i
            jj = len(parameters) - j - 1

            xmajorLocator = matplotlib.ticker.MaxNLocator(nbins=4,prune='both')
            ymajorLocator = matplotlib.ticker.MaxNLocator(nbins=4,prune='both')

            if j <= len(parameters)-i-1:
                axarr[jj][ii].xaxis.set_minor_locator(NullLocator())
                axarr[jj][ii].yaxis.set_minor_locator(NullLocator())
                axarr[jj][ii].xaxis.set_major_locator(NullLocator())
                axarr[jj][ii].yaxis.set_major_locator(NullLocator())

                axarr[jj][ii].xaxis.set_minor_formatter(NullFormatter())
                axarr[jj][ii].yaxis.set_minor_formatter(NullFormatter())
                axarr[jj][ii].xaxis.set_major_formatter(NullFormatter())
                axarr[jj][ii].yaxis.set_major_formatter(NullFormatter())
                xmajorFormatter = FormatStrFormatter('%g')
                ymajorFormatter = FormatStrFormatter('%g')

                if ii == jj:
                    # Make a 1D plot
                    makesubplot1d(axarr[ii][ii], chain[:,parameters[ii]], \
                                  weights=weights, interpolate=interpolate, \
                                  smooth=smooth)
                    axarr[ii][jj].set_ylim(ymin=0)

                    if inj is not None:
                        axarr[ii][ii].axvline(inj[ii], lw=2, color='k')
                else:
                    # Make a 2D plot
                    makesubplot2d(axarr[jj][ii], chain[:,parameters[ii]], \
                            chain[:,parameters[jj]],color=color, weights=weights, \
                                  smooth=smooth)

                    if inj is not None:
                        axarr[jj][ii].plot(inj[ii], inj[jj], 'x', color='k', markersize=12, \
                                           mew=2, mec='k')

                axarr[jj][ii].xaxis.set_major_locator(xmajorLocator)
                axarr[jj][ii].yaxis.set_major_locator(ymajorLocator)
            else:
                axarr[jj][ii].set_visible(False)
                #axarr[jj][ii].axis('off')

            if jj == len(parameters)-1:
                axarr[jj][ii].xaxis.set_major_formatter(xmajorFormatter)
                if labels:
                    axarr[jj][ii].set_xlabel(labels[ii])

            if ii == 0:
                if jj == 0:
                    axarr[jj][ii].yaxis.set_major_locator(NullLocator())
                    #axarr[jj][ii].set_ylabel('Post.')
                else:
                    axarr[jj][ii].yaxis.set_major_formatter(ymajorFormatter)
                    if labels:
                        axarr[jj][ii].set_ylabel(labels[jj])

    # overall plot title
    if title:
        f.suptitle(title, fontsize=14, y=0.90)
     
    # make plots closer together 
    f.subplots_adjust(hspace=0.1)
    f.subplots_adjust(wspace=0.1)


def pol2cart(lon, lat): 
    """ 
    Utility function to convert longitude,latitude on a unit sphere to 
    cartesian co-ordinates.

    """ 

    x = np.cos(lat)*np.cos(lon) 
    y = np.cos(lat)*np.sin(lon) 
    z = np.sin(lat) 

    return np.array([x,y,z])


def greedy_bin_sky(skypos, skycarts):
    """

    Greedy binning algorithm

    """

    N = len(skycarts) 
    skycarts = np.array(skycarts)
    bins = np.zeros(N) 
    for raSample, decSample in skypos: 
        sampcart = pol2cart(raSample, decSample) 
        dx = np.dot(skycarts, sampcart)
        maxdx = np.argmax(dx)
        bins[maxdx] += 1 

    # fill in skymap
    histIndices = np.argsort(bins)[::-1]    # in decreasing order
    NSamples = len(skypos)

    frac = 0.0
    skymap = np.zeros(N)
    for i in histIndices:
        frac = float(bins[i])/float(NSamples) 
        skymap[i] = frac

    return skymap


def plotSkyMap(raSample, decSample, nside=64, contours=None, colorbar=True, \
              inj=None, psrs=None):
    """

    Plot Skymap of chain samples on Mollwiede projection.

    @param raSample: Array of right ascension samples
    @param decSample: Array of declination  samples
    @param nside: Number of pixels across equator [default = 64]
    @param contours: Confidence contours to draw eg. 68%, 95% etc
                     By default this is set to none and no contours
                     will be drawn.
    @param colorbar: Boolean option to draw colorbar [default = True]
    @param inj: list of injected values [ra, dec] in radians to plot
                [default = None]
    @param psrs: Stacked array of pulsar sky locations [ra, dec] in radians
                 [default=None] Will plot as white diamonds

    """

    # clear figures
    plt.clf()

    # create stacked array of ra and dec
    skypos = np.column_stack([raSample, decSample])

    npix = hp.nside2npix(nside)    # number of pixels total


    # initialize theta and phi map coordinantes
    skycarts=[]
    for ii in range(npix):
        skycarts.append(np.array(hp.pix2vec(nside,ii)))

    # get skymap values from greedy binning algorithm
    skymap = greedy_bin_sky(skypos, skycarts)

    # smooth skymap
    skymap = hp.smoothing(skymap, 0.05)

    # make plot
    ax = plt.subplot(111, projection='astro mollweide')

    # Add contours
    if contours is not None:
        for percent in contours:
            indices = np.argsort(-skymap)
            sky = skymap[indices]
            region = np.zeros(skymap.shape)
            ind = np.min(ml.find(np.cumsum(sky) >= 0.01*percent))
            region[indices[0:ind]] = 1.0
            cs = bplot.contour(lambda lon, lat: region[hp.ang2pix(nside, 0.5*np.pi - lat, lon)], \
                          colors='k', linewidths=1.0, levels=[0.5])
            #plt.clabel(cs, [0.5], fmt={0.5: '$\mathbf{%d\%%}$' % percent}, fontsize=8, inline=True)

    # plot map
    ax.grid()
    bplot.outline_text(ax)
    bplot.healpix_heatmap(skymap)

    # add injection
    if inj:
        ax.plot(inj[0], inj[1], 'x', color='k', markersize=8, mew=2, mec='k')

    # add pulsars
    if np.all(psrs):
        ax.plot(psrs[:,0], psrs[:,1], 'D', color='w', markersize=3, mew=1, mec='w')

    # add colorbar and title
    if colorbar:
        plt.colorbar(orientation='horizontal')
        plt.suptitle(r'$p(\alpha,\delta|d)$', y=0.1)

    # save skymap
    plt.savefig('skymap.pdf', bbox_inches='tight')



# def upperlimitplot2d(x, y, sigma=0.95, ymin=None, ymax=None, bins=40, log=False, \
#                      savename=None, labels=None, hold=False, **kwargs):

#     """

#     Make upper limits of a parameter as a function of another.

#     @param x: Parameter we are making upper limits for
#     @param y: Parameter which we will bin
#     @param sigma: Sigma level of upper limit
#     @param ymin: Minimum value of binning parameter [default=None]
#     @param ymax: Maximum value of binning parameter [default=None]
#     @param bins: Number of bins
#     @param log: If True, plot on log-log scale
#     @param savename: Output filename for saved figure
#     @param labels: List of labels for axes [xlabel, ylabel]
#     @param hold: Hold current figure?

#     """

#     # clear current figure
#     if hold == False:
#         plt.clf()

#     if ymin is None:
#         ymin = y.min()
#     if ymax is None:
#         ymax = y.max()

#     yedges = np.linspace(ymin, ymax, bins+1)
#     deltay = yedges[1] - yedges[0]
#     yvals = np.linspace(ymin+0.5*deltay, ymax-0.5*deltay, bins)
#     bin_index = []
#     upper = []

#     for i in range(bins):
#         # Obtain the indices in the range of the bin
#         indices = np.flatnonzero(np.logical_and(y>yedges[i], y<yedges[i+1]))

#         # Obtain the 1-sided x-sigma upper limit
#         if len(indices) > 0:
#             bin_index.append(i)
#             a, sigma1 = confinterval(x[indices], sigma=sigma, onesided=True)
#             upper.append(sigma1)

#     # make bin_indes and upper into arrays
#     bin_index = np.array(bin_index)
#     upper = np.array(upper)

#     # make plot
#     if log:
#         plt.loglog(10**yvals[bin_index], 10**upper, **kwargs)
#         plt.grid(which='major')
#         plt.grid(which='minor')
#     else:
#         plt.plot(yvals[bin_index], upper, **kwargs)
#         plt.grid()

#     # labels
#     if labels:
#         plt.xlabel(labels[0])
#         plt.ylabel(labels[1])

#     if savename:
#         plt.savefig(savename, bbox_inches='tight')
#     else:
#         plt.savefig('2dUpperLimit.pdf', bbox_inches='tight')

    




import numpy as np
from matplotlib import pylab as plt
import scipy

#import mcsampler
import RIFT.integrators.mcsamplerGPU as mcsampler
import ourio


# Specify a fixed parameter
samplerPrior = mcsampler.MCSampler()
samplerPrior.add_parameter('x', np.vectorize(lambda x: 1.), None, 0,1,prior_pdf=np.vectorize(lambda x:1.))  # is this correctly renormalized?
# Do an MC integral with this sampler (=the measure specified by the sampler).
res, var,  neff, dict_return = samplerPrior.integrate(np.vectorize(lambda x:1.0), 'x', nmax=1e4,verbose=True,args='x')
print(("Integral of 1 over this range ", [samplerPrior.llim['x'] ,samplerPrior.rlim['x'] ], " is ", res, " needs to be ", samplerPrior.rlim['x'] -  samplerPrior.llim['x']))
print( "     -> no adaptivity used ")


# Specify an integral, use adaptivity
samplerPrior = mcsampler.MCSampler()
samplerPrior.add_parameter('x', np.vectorize(lambda x: 1.), None, 0,1,prior_pdf=np.vectorize(lambda x:1.),adaptive_sampling=True)  # is this correctly renormalized?
# Do an MC integral with this sampler (=the measure specified by the sampler).
sigma_val=0.5
res, var,  neff, dict_return = samplerPrior.integrate( lambda x: np.exp(-(x/sigma_val)**2/2.)/np.sqrt(2*np.pi*sigma_val**2), 'x', nmax=1e4,verbose=True,args='x')
print(("Integral of 1 over this range ", [samplerPrior.llim['x'] ,samplerPrior.rlim['x'] ], " is ", res, " needs to be ", scipy.stats.norm.cdf(samplerPrior.rlim['x']/sigma_val) -  scipy.stats.norm.cdf(samplerPrior.llim['x']/sigma_val)))




# do an integral with a different sampling prior
samplerNewPrior = mcsampler.MCSampler()
samplerNewPrior.add_parameter('x', np.vectorize(lambda x: np.exp(-x**2/2.)/np.sqrt(2*np.pi)), None, -1,1, prior_pdf = np.vectorize(lambda x: 1/2.)) # normalized prior!
res, var,  neff, dict_return = samplerNewPrior.integrate(np.vectorize(lambda x:1), 'x', nmax=1e4,verbose=True)
print(("Integral of 1 using normal sampling is  ", res, " and needs to be 1"))


# Manual plotting. Note the PDF is NOT renormalized
try:
    xvals = np.linspace(-1,1,100)
    yvals = samplerPrior.pdf['x'](xvals)
    ypvals = samplerPrior.pdf['x'](xvals)/samplerPrior._pdf_norm['x']
    zvals = samplerPrior.cdf['x'](xvals)
    plt.plot(xvals, yvals,label='pdf shape')
    plt.plot(xvals, ypvals,label='pdf')
    plt.plot(xvals, zvals,label='cdf')
    plt.ylim(0,2)
    plt.legend()
    plt.show()
except:
    print(" Alternate interfae does not support raw access to priors ")

try:
    # Automated plotting 
    sampler = mcsampler.MCSampler()
    sampler.add_parameter('x', np.vectorize(lambda x: np.exp(-x**2/2)), None, -1,1)
    ourio.plotParameterDistributions('sampler-foridiots-example', sampler)
except:
    print(" Alternate interface does not support alternate plots")

#print "Integral of 1 over this range ", [samplerPrior.llim['y'] ,samplerPrior.rlim['y'] ], " is ", ret, " needs to be 1, because I used a normalized prior"



# Do an MC integral and return the sampling points
# Stop after I reach neff = 100 points (should be fast!)
# sig = 0.1
# print(" -- Performing integral, stopping after neff = 100 points -- ")
# res, var, ret, lnLmarg, neff = samplerPrior.integrate(np.vectorize(lambda x: np.exp(-x**2/(2*sig**2))), 'x', nmax=5*1e4, full_output=True,neff=100)
# print(" integral answer is ", res,  " with expected error ", np.sqrt(var), ";  compare to ", np.sqrt(2*np.pi)*sig)
# print(" note neff is ", neff, "; compare neff^(-1/2) = ", 1/np.sqrt(neff), " to relative predicted and actual errors: ", np.sqrt(var)/res, ", ",  (res - np.sqrt(2*np.pi)*sig)/res)
# # Save the sampled points to a file
# ourio.dumpSamplesToFile("sampler-foridiots-example.dat", ret, ['x', 'lnL'])
# # plot the array of partial sums, to illustrate convergence
# print(" -- Integral convergence ---")
# xarr, lnL = np.transpose(ret)
# lnLmax = np.maximum.accumulate(lnL)
# plt.plot(np.arange(len(lnLmax)), lnLmax,label='lnLmax')      # array of running-maximum lnL values
# plt.plot(np.arange(len(lnLmarg)), lnLmarg,label='lnLmarg')    # array of marginalized L values
# plt.xlabel('iteration')
# plt.ylabel('lnL')
# plt.legend()
# plt.show()
# plt.clf()
# # Plot the sampled points (as example for direct access to 'ret')
# print(" -- Sampled array of points  -- ")
# xvals = np.transpose(ret)[0]
# yvals = np.exp(np.transpose(ret)[1])  # lnL is last element of array
# plt.scatter(xvals,yvals)
# plt.xlabel('x')
# plt.ylabel('L')
# plt.show()





import numpy as np
import scipy.integrate

xpy_default=np
has_cupy = False
try:
    import cupy   # can proceed even if cupy doesn't actually work
    junk_to_check_installed = cupy.array(5)
    xpy_default=cupy
    has_cupy=True
except:
    has_cupy=False  # just to be sure
# try:
#     import numba
#     from numba import vectorize, complex128, float64, int64
#     numba_on = True
#     print(" Numba on (priors_utils) ")
    

# except:
#     numba_on = False
#     print(" Numba off (priors_utils) ")

# https://git.ligo.org/RatesAndPopulations/lalinfsamplereweighting/blob/reviewed-post-O2/approxprior.py
will_cosmo_const = np.array( [ 1.012306, 1.136740, 0.262462, 0.016732, 0.000387 ])
p_in = will_cosmo_const[::-1]
if has_cupy:
    p_in = cupy.asarray(p_in)
def dist_prior_pseudo_cosmo(dL,nm=1,xpy=np,p_in=p_in):
    """
    dist_prior_pseudo_cosmo.  dL needs to be in Gpc for the polynomial.
    By default, our code works with distances in Mpc.  So we divide by 1e3

     Will Farr's simplified distance prior on d_L, out to z~ 4
     note it is not normalized, and the normalization depends on the d_max of interest 
    
    """
    if isinstance(dL, np.ndarray):
        return nm*4* np.pi * dL**2 / np.polyval( will_cosmo_const[::-1], dL/1e3)
    return nm*4* np.pi * dL**2 / xpy.polyval( p_in,dL/1e3)


def dist_prior_pseudo_cosmo_eval_norm(dLmin,dLmax):
    return 1./scipy.integrate.quad(dist_prior_pseudo_cosmo, dLmin,dLmax)[0]


## VARIOUS COSMOLOGY THINGS

def get_astropy_cosmology(name="Planck15"):
   from astropy import cosmology as cosmo
   return getattr(cosmo, name)


## INTERPOLATED PRIORS: General tools (target: cosmology/ILE)

def norm_cdf_and_inverse(dPdx, xlim=[0,1],vectorized=True,rtol=1e-6,method='DOP853',**kwargs):
  """
  Finds forward and inverse CDF via accurate ODE solve.  Makes sure to correct for normalization.
  Danger: be careful if integrand goes to zero before the left limit!
  Targeting reasonably high tolerance (of order 1e-6)

  Weakness: pure CPU return value.  Wrap in interpolating function call to force use of returning GPU-compatible function
     Could try https://github.com/yor-dev/cupy_ivp/tree/master
  """
  nm = scipy.integrate.quad(dPdx, xlim[0], xlim[1])[0]
 # print(nm)
  def func_inv(P,x):
    return 1/dPdx(x)
  def func_forward(x,P):
    return dPdx(x)
  sol_inv = scipy.integrate.solve_ivp(func_inv,(0,nm),[xlim[0]],vectorized=vectorized, dense_output=True,rtol=rtol,method=method, **kwargs)
  if not(sol_inv.success):
    print(sol_inv)
    raise Exception("Prior: normalization failure")
#  print(sol_inv)
  sol_forward = scipy.integrate.solve_ivp(func_forward,(xlim[0],xlim[1]),[0],vectorized=vectorized, dense_output=True,atol=1e-8,rtol=rtol,method=method, **kwargs)
  if not(sol_forward.success):
    print(sol_forward)
    raise Exception("Prior: normalization failure")
#  print(sol_forward)
  return nm, lambda x,f=sol_inv.sol, c=nm: f(x*c)[0],lambda x,f=sol_forward.sol, c=nm: f(x)[0]/c

def norm_and_inverse_via_grid_interp(dPdx, xlim=[1e-5,1], y_of_x=None, loglog=False,log_grid=True,interp_method_name='PchipInterpolator' ,final_scipy_interpolate=scipy.interpolate,final_np=np,to_gpu=lambda x: x, to_gpu_needed=False,n_grid=1000, **kwargs):
   interp_action = getattr(final_scipy_interpolate, interp_method_name)
   nm, sol_inv, sol_forward = norm_cdf_and_inverse(dPdx, xlim, **kwargs)
   if log_grid:
    # grid 1, for logarithmic spacing for interpolation of dp/dx, NOT UNIFORM. PROBLEM: linear interpolation sub-optimal on this dynamic range.
    x_vals = np.exp(np.linspace(np.log(xlim[0]), np.log(xlim[1]), num=n_grid))
   else:
    x_vals = np.linspace(xlim[0], xlim[1], num=n_grid)
   dp_vals = dPdx(x_vals)/nm # remove scale factor from dP/dx
   P_vals = sol_forward(x_vals)
   P_vals *= 1./P_vals[-1] # rescale to CDF unity at right limit. Normalization should be 1 from previous code
   dp_vals *= 1./P_vals[-1] # similarly, scale out normalization constant. Normalizatio should be 1 from previous code
   if y_of_x:
    x_vals = y_of_x(x_vals) # reparameterize
    if loglog:
      raise Exception(" interpolation of cdf/pdf: reparameterization dP/dy not enabled with log log")
   # convert to GPU if needed
   if to_gpu_needed:
    x_vals = to_gpu(x_vals)
    dp_vals = to_gpu(dp_vals)
    P_vals = to_gpu(P_vals)
   if not(loglog):
      # interpolation with default method
      dp_func = interp_action(x_vals, dp_vals)
      cdf_func = interp_action(x_vals, P_vals)
      cdf_inv_func = interp_action(P_vals, x_vals)
      # reparameterize action completed for PDF
      if y_of_x:
        dp_func = cdf_func.derivative()
      return dp_func, cdf_func, cdf_inv_func
   # interpolation with default method, using log-log variables for dynamic range stability
   #.  - warning: dp/dy not implemented
   #print(np.log(x_vals))
   dp_func_log = interp_action(final_np.log(x_vals), dp_vals) # dp(ln x). Incorrect if we have changed to y
   cdf_vals_func_log_log = interp_action(final_np.log(x_vals),final_np.log(P_vals)) # lnP (lnx)
   cdf_inv_func_log_log =interp_action(final_np.log(P_vals),final_np.log(x_vals)) # lnx (lnP)
   # wrap in functions
   cdf_func_wrapper = lambda x: final_np.exp(cdf_vals_func_log_log(final_np.log(x)))
   cdf_inv_func_wrapper = lambda x: final_np.exp(cdf_inv_func_log_log(final_np.log(x)))
   pdf_func_wrapper = lambda x: dp_func_log(final_np.log(x))
   # return
   return pdf_func_wrapper, cdf_func_wrapper, cdf_inv_func_wrapper



#!/usr/bin/env python
#
#  GOAL
#     - demonstrate spoke reconstruction from xml and from ILE *.dat output
#
#
# EXAMPLE
#    - util_TestSpokesIO.py --fname overlap-grid.xml.gz  --fname-dat ~/g184098-FlipFlop.dat --test-refinement
#    - util_TestSpokesIO.py   --fname-dat ~/g184098-RIT-Generic-v0dfix.dat  --verbose --test-refinement
# EXAMPLE FULL REFINEMENT
#   python util_TestSpokesIO.py   --fname-dat ~/g184098-FlipFlop.dat  --fname overlap-grid.xml.gz   --test-refinement
import RIFT.misc.spokes as spokes

import argparse
import sys
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lal

try:
        from matplotlib import pyplot as plt
except:
        print(" no plots")
###
### Load options
###

parser = argparse.ArgumentParser()
parser.add_argument("--fname", default=None, help="Base output file for XML loading")
parser.add_argument("--fname-dat",default=None)
parser.add_argument("--is-eccentric",action='store_true',default=False)
parser.add_argument("--test-refinement",action='store_true')
parser.add_argument("--save-refinement-fname", default="refined-grid",help="Output to save refined grid xml. REQUIRES valid pairing")
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--mega-verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
opts=  parser.parse_args()



###
### Load xml grid
###

if opts.fname and opts.mega_verbose:
 sdHere = spokes.LoadSpokeXML(opts.fname,is_eccentric=opts.is_eccentric)
 for key in sdHere:
    print(" Spoke : ", key)
    for P in sdHere[key]:
        mtot = spokes.ChooseWaveformParams_to_spoke_mass(P)  # used so rounding is consistent
        print(mtot)


###
### Load ILE .dat output.  REFINE test
###

if opts.fname_dat:
 sdHere = spokes.LoadSpokeDAT(opts.fname_dat,is_eccentric=opts.is_eccentric)
 print(" Spoke count : ", len(sdHere.keys()))
 fig_index =0
 for key in sdHere:
    sdHereCleaned = spokes.CleanSpokeEntries(sdHere[key])
    print(" Spoke ", key, len(sdHereCleaned))
    if opts.mega_verbose:
     for spoke_entry in sdHere[key]:
        print(spoke_entry)
     print(" --- cleaning --- ")
     for spoke_entry in sdHereCleaned:
        print(spoke_entry)
    if opts.test_refinement:
        #print sdHereCleaned[:,0], sdHereCleaned[:,1]
        code, xvals_new = spokes.Refine(sdHereCleaned[:,0], sdHereCleaned[:,1],xmin=1)  
        if xvals_new is None:
                continue
        xvals_new = np.array(xvals_new)
        xvals_new = xvals_new[ xvals_new > 0]  # eliminate negtive masses!

        if opts.mega_verbose:
           print(code, xvals_new)

    if opts.verbose:
        fig_index+=1
        plt.figure(fig_index)
#        plt.plot(sdHereCleaned[:,0], sdHereCleaned[:,1],'o')
        plt.text(np.mean(sdHereCleaned[:,0])-10,np.max(sdHereCleaned[:,1]), str(key))
        plt.errorbar(sdHereCleaned[:,0], sdHereCleaned[:,1],yerr=sdHereCleaned[:,2],linestyle='none')
        if opts.test_refinement and code!='fail':
            plt.plot(xvals_new, 200*np.ones(len(xvals_new)),'o')

        xvals_min_here = np.min(sdHereCleaned[:,0])
        xvals_max_here = np.max(sdHereCleaned[:,0])
        fn = spokes.FitSpokeNearPeak(sdHereCleaned[:,0], sdHereCleaned[:,1])
        xvals_plot = np.linspace(xvals_min_here,xvals_max_here,100)
        plt.plot(xvals_plot,fn(xvals_plot),'k')

if opts.verbose:
    plt.show()


###
### Use DAT + XML lookup to generate refinement, per spoke
###


if opts.fname and opts.fname_dat:
   sd_dat = spokes.LoadSpokeDAT(opts.fname_dat,is_eccentric=opts.is_eccentric)
   sd_P =  sdHere = spokes.LoadSpokeXML(opts.fname,is_eccentric=opts.is_eccentric)

   print(" --- Refinement: Writing XML --- ")
   print(" data file ", len(sd_dat))
   print(" xml file ", len(sd_P))

   P_list = []
   nCount = 0;
   nFailures = 0
   for spoke_id in sd_dat.keys():
      nCount +=1
      # Cross-look-up
      print(spoke_id)
      try:
         P_sample = sd_P[spoke_id][0] # if this fails
         P_sample.waveFlags =None
         P_sample.nonGRparams = None
         P_sample.print_params()
      except:
         nFailures +=1
         print(" Failed cross lookup for ", spoke_id, nCount, " failure count = ", nFailures)
         continue
      # Clean
      sd_here =spokes.CleanSpokeEntries(sd_dat[spoke_id])
      # Refine: find mass values
      code, mvals_new = spokes.Refine(sd_here[:,0], sd_here[:,1])
      if mvals_new is None:
              continue  # Failed
      mvals_new = np.array(mvals_new)
      mvals_new = mvals_new[ mvals_new > 0]  # eliminate negtive masses!
      print(key, len(sd_here), code, mvals_new)
      if code == 'refined' or code =='extended':
         for m in mvals_new:
            print(m)
            P = P_sample.manual_copy()
            P.tref = P_sample.tref   # BE VERY CAREFUL: The structure swig-bound to store time is NOT a float, and requires careful memory management
            P.assign_param('mtot',m*lal.MSUN_SI)
            P_list.append(P)
            
   if P_list:
      lalsimutils.ChooseWaveformParams_array_to_xml(P_list, opts.save_refinement_fname)



#! /usr/bin/env python
#
#  GOAL
#   Iterate CIP until convergence
#     * Create many stages of calling a subsag creating job


import argparse
import sys
import os
import shutil
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lal
import functools
import itertools

from glue import pipeline # https://github.com/lscsoft/lalsuite-archive/blob/5a47239a877032e93b1ca34445640360d6c3c990/glue/glue/pipeline.py

import RIFT.misc.dag_utils as dag_utils
from RIFT.misc.dag_utils import mkdir
from RIFT.misc.dag_utils import which


parser = argparse.ArgumentParser()
parser.add_argument("--working-directory",default="./")
parser.add_argument("--n-eff-goal",default=10000,type=int)
parser.add_argument("--n-eff-per-worker",default=200,type=int,help="n-eff goal per worker CIP ")
parser.add_argument("--cip-exe",default=None,help="filename of CIP or equivalent executable. Will default to `which util_ConstructIntrinsicPosterior_GenericCoordinates` in low-level code")
parser.add_argument("--cip-args",default=None,help="filename of args_cip.txt file  which holds CIP arguments.  Should NOT conflict with arguments auto-set by this DAG ... in particular, i/o arguments will be modified.   We will NOT change --n-max, so user is assumed to have set this intelligently before trying this.  ASSUME INITIAL ITERATION COUNT ALREADY STRIPPED")
parser.add_argument('--n-iterations',default=10,type=int,help="Maximum number of iterations to perform before terminating")
parser.add_argument("--cip-explode-jobs-min",default=3,type=int,help="Min number of jobs to use in parallel.  ")
parser.add_argument("--cip-explode-jobs-max",default=3,type=int,help="Min number of jobs to use in parallel.  ")
parser.add_argument("--transfer-file-list",default=None,help="File containing list of *input* filenames to transfer, one name per file. Copied into transfer_files for condor directly.  If provided, also enables attempts to deduce files that need to be transferred for the pipeline to operate, as needed for OSG, etc")
parser.add_argument("--request-memory-CIP",default=16384,type=int,help="Memory request for condor (in Mb) for fitting jobs.")
parser.add_argument("--use-singularity",action='store_true',help="Attempts to use a singularity image in SINGULARITY_RIFT_IMAGE")
parser.add_argument("--use-osg",action='store_true',help="Attempts to set up an OSG workflow.  Must submit from an osg allowed submit machine")
parser.add_argument("--use-osg-simple-requirements",action='store_true',help="Uses an aggressive simplified requirements string for worker jobs")
parser.add_argument("--condor-local-nonworker",action='store_true',help="Uses local universe for non-worker condor jobs. Important to run in non-NFS location, as other jobs don't have file transfer set up.")
parser.add_argument("--condor-nogrid-nonworker",action='store_true',help="Uses local flocking for non-worker condor jobs. Important to run in non-NFS location, as other jobs don't have file transfer set up.")
parser.add_argument("--general-retries",default=1,type=int,help="Number of retry attempts for internal jobs (convert, CIP, ...). (These can fail, albeit more rarely, usually due to filesystem problems)")
parser.add_argument("--general-request-disk",default="4M",type=str,help="Request disk passed to condor. Must be done for all jobs now")
parser.add_argument("--use-full-submit-paths",action='store_true',help="DAG created has full paths to submit files generated. Note this is implemented on a per-file/as-needed basis, mainly to facilitate using this dag as an external subdag")
opts=  parser.parse_args()


local_worker_universe="vanilla"
no_worker_grid=False
if opts.condor_local_nonworker:
    local_worker_universe="local"
if opts.condor_nogrid_nonworker:
    no_worker_grid=True


working_dir_inside_local = working_dir_inside = opts.working_directory
os.chdir(opts.working_directory)
# working_dir_inside_local is for jobs on local nodes
if opts.use_singularity or opts.use_osg:
    working_dir_inside = "./" # all files on the remote machine are in the current directory

singularity_image = None
if opts.use_singularity:
    print(" === USING SINGULARITY === ")
    singularity_image = os.environ["SINGULARITY_RIFT_IMAGE"]  # must be present to use singularity
    # SINGULARITY IMAGES ARE ON CVMFS, SO WE CAN AVOID THE SINGULARITY EXEC CALL
    # hardcoding a fiducial copy of lalapps_path2cache; beware about the executable name change
    os.environ['LALAPPS_PATH2CACHE'] = "/cvmfs/oasis.opensciencegrid.org/ligo/sw/conda/envs/igwn-py39/bin/lalapps_path2cache" #"singularity exec {singularity_image} lalapps_path2cache".format(singularity_image=singularity_image)
    print(singularity_image)

if (opts.cip_args is None):
    print(" No arguments provided for low-level job")
    sys.exit(0)

if True:
    with open(opts.cip_args) as f:
        cip_args_list = f.readlines()
    cip_args = ' '.join( [x.replace('\\','') for x in cip_args_list] )
    cip_args = ' '.join(cip_args.split(' ')[1:])
    # Some argument protection for later
    cip_args = cip_args.replace('[', ' \'[')
    cip_args = cip_args.replace(']', ']\'')
    cip_args=cip_args.rstrip()
    cip_args += ' --no-plots '  

print("CIP", cip_args)

transfer_file_names = []
if not (opts.transfer_file_list is None):
    transfer_file_names=[]
    # Load args.txt. Remove first item.  Store
    with open(opts.transfer_file_list) as f:
        for  line in f.readlines():
            transfer_file_names.append(line.rstrip())
    print(" Input files to transfer to job working directory (note!)", transfer_file_names)



###
### DAG generation
###

dag = pipeline.CondorDAG(log=os.getcwd())

# Make directories for all iterations
mkdir(opts.working_directory+"/logs")



##
## Command used for assessment job (except for iteration number) will always be the same
cmd_subdag = "  --iteration-number $(macroiteration)  --n-iterations {}".format(opts.n_iterations)
cmd_subdag += " --cip-exe {} --cip-args {} ".format(opts.cip_exe,opts.cip_args)
cmd_subdag += " --n-eff-goal {} --n-eff-per-worker {} ".format(opts.n_eff_goal, opts.n_eff_per_worker)
cmd_subdag += " --cip-explode-jobs-min {} --cip-explode-jobs-max {} ".format(opts.cip_explode_jobs_min,opts.cip_explode_jobs_max)
cmd_subdag += " --general-retries {} --general-request-disk {} --request-memory-CIP  ".format(opts.general_retries,opts.general_request_disk,opts.request_memory_CIP)
if opts.transfer_file_list:
    cmd_subdag += " --transfer-file-list {} ".format(opts.transfer_file_list)
if opts.use_osg:
    cmd_subdag += " --use-osg "
if opts.use_osg_simple_requirements:
    cmd_subdag += " --use-osg-simple-requirements "
if opts.use_singularity:
    cmd_subdag += " --use-singularity "
if opts.condor_local_nonworker:
    cmd_subdag += " --condor-local-nonworker "
if opts.condor_nogrid_nonworker:
    cmd_subdag += " --condor-nogrid-nonworker "
print(" Iteration command arguments:  ",cmd_subdag)

# Write a submit file which creates the subdag needed
#  REUSE the generic 'con' submit from misc.dag_utils
con_job, con_job_name = dag_utils.write_consolidate_sub_simple(tag='assess_propose',exe='assess_propose_CIP_convergence_sequence.py',log_dir=None,arg_str='',base=opts.working_directory, target='',universe=local_worker_universe,no_grid=no_worker_grid)
#print(dir(con_job))
con_job._CondorJob__arguments.remove(opts.working_directory)
con_job.add_arg(cmd_subdag)
# Modify: set 'initialdir' : should run in top-level direcory
#con_job.add_condor_cmd("initialdir",opts.working_directory+"/iteration_$(macroiteration)_con")
# Modify output argument: change logs and working directory to be subdirectory for the run
con_job.set_log_file(opts.working_directory+"//logs/ap-$(macroiteration)-$(cluster)-$(process).log")
con_job.set_stderr_file(opts.working_directory+"/logs/ap-$(macroiteration)-$(cluster)-$(process).err")
con_job.set_stdout_file(opts.working_directory+"//logs/ap-$(macroiteration)-$(cluster)-$(process).out")
con_job.add_condor_cmd('request_disk',opts.general_request_disk)
if opts.use_full_submit_paths:
    fname = opts.working_directory+"/"+con_job.get_sub_file()
    con_job.set_sub_file(fname)
con_job.write_sub_file()




# Write subdag placeholders
parent_fit_node = None


for it in np.arange(0,opts.n_iterations):
    ###
    ### ASSESSMENT JOB
    con_node = pipeline.CondorDAGNode(con_job)
    con_node.add_macro("macroiteration",it)
    con_node.set_retry(opts.general_retries)
    if (parent_fit_node):
        con_node.add_parent(parent_fit_node)

    ###
    ### WORKING FROM ASSESSMENT JOB
    main_subdag =  pipeline.CondorDAGManJob("{}/convergence_cip_{}_subdag.dag".format(opts.working_directory,it))  # assumes subdag is created
    main_analysis_node = main_subdag.create_node()
    main_analysis_node.set_retry(opts.general_retries)
    main_analysis_node.add_parent(con_node)

    parent_fit_node = main_analysis_node
    dag.add_node(con_node)
    dag.add_node(main_analysis_node)
    


dag_name="cip_converge_workflow"
dag.set_dag_file(dag_name)
dag.write_concrete_dag()
